==> quantagonia-0.1.11-py3-none-any.whl: quantagonia/cloud/https_client.py <==
from __future__ import annotations

import copy
import json
import os
import uuid
from typing import Tuple

import requests
from requests.adapters import HTTPAdapter
from urllib3 import Retry

from quantagonia.cloud.dto.encoder import DtoJSONEncoder
from quantagonia.cloud.dto.job import JobDto
from quantagonia.cloud.dto.presigned_s3 import PresignedS3
from quantagonia.cloud.enums import Endpoints
from quantagonia.enums import HybridSolverServers
from quantagonia.errors import SolverError


class HTTPSClient:
    """HTTPS Client for the HybridSolver with low-level functions to manage optimization tasks.

    Args:
        api_key (str): The API key to be used to authenticate with the HybridSolver.
        server (HybridSolverServers): (optional) Allows to specify a server different to the default production server.
                                      Defaults to :code:`None`.
        custom_headers (dict): (optional) Custom headers for the https requests.

    """

    def __init__(self, api_key: str, server: HybridSolverServers = None, custom_headers: dict = None) -> None:
        self.api_key = api_key
        if server is None:
            # Read the default server from an env variable. If none is set, the server defaults to PROD
            server = HybridSolverServers[os.environ.get("QUANTAGONIA_SERVER", "PROD").upper()]
        self.server = server.value
        self.custom_headers = custom_headers if custom_headers is not None else {}

        self.retry = Retry(total=5, backoff_factor=1, allowed_methods=frozenset(["GET", "POST", "PUT", "DELETE"]))
        self.session = requests.Session()
        self.session.mount("http://", HTTPAdapter(max_retries=self.retry))
        self.session.mount("https://", HTTPAdapter(max_retries=self.retry))

    def submit_job(self, problem_files: list, specs: list, tag: str = "", context: str = "") -> str:
        # build a single JSON array with specs
        spec_arr = {}
        for ix in range(0, len(specs)):
            spec_arr[str(ix)] = specs[ix]

        job_id = str(uuid.uuid4())
        problem_filenames = []
        for ind, prob_file in enumerate(problem_files):
            try:
                filename = prob_file.parts[len(prob_file.parts) - 1]
            except AttributeError:
                filename = prob_file.split("/")[len(prob_file.split("/")) - 1]

            problem_filenames.append(filename)
            ps3_problem_file = PresignedS3(
                jobId=job_id, contentType="application/octet-stream", batchNumber=str(ind), fileName=filename
            )
            data_for_s3 = json.dumps(ps3_problem_file, cls=DtoJSONEncoder)
            response = self.session.post(
                self.server + Endpoints.s3,
                data=data_for_s3,
                headers={"X-api-key": self.api_key, "Content-type": "application/json", **self.custom_headers},
            )
            if not response.ok:
                msg = "Unable to get an S3 presigned URL"
                raise RuntimeError(msg)

            file_uploader = response.json()
            headers = {
                "X-amz-meta-author": file_uploader["metaAuthor"],
                "X-amz-meta-version": file_uploader["metaVersion"],
                "Content-type": "application/octet-stream",
            }
            with open(prob_file, "rb") as f:
                upload_response = self.session.put(file_uploader["url"], data=f, headers=headers)

            if not upload_response.ok:
                raise RuntimeError(upload_response.text)

        start_job = JobDto(jobId=job_id, problemFiles=problem_filenames, specs=specs, tag=tag, context=context)
        start_job_data = json.dumps(start_job, cls=DtoJSONEncoder)
        started_job = self.session.post(
            self.server + Endpoints.job,
            data=start_job_data,
            headers={"X-api-key": self.api_key, "Content-type": "application/json", **self.custom_headers},
        )
        if not started_job.ok:
            error_report = started_job.json()
            raise RuntimeError(error_report)

        return started_job.json()["jobId"]

    def _replace_file_content_from_url(self, e: dict, key: str) -> None:
        e[key] = self._get_file_content_from_url(e[key])

    def _get_file_content_from_url(self, e: dict | str) -> str:
        if isinstance(e, dict) and "error" in e:
            return "Error: " + e["error"]
        if isinstance(e, dict) and "url" in e:
            if e["url"] == "":
                return ""
            response = self.session.get(e["url"])
            if response.status_code == 200:
                return response.text
            return ""
        return e

    def check_job(self, jobid: str) -> str:
        params = {"jobid": str(jobid)}
        response = self.session.get(
            self.server + Endpoints.checkjob, params=params, headers={"X-api-key": self.api_key, **self.custom_headers}
        )
        if response.ok:
            return response.json()["status"]
        if response.status_code > 499:
            log = self.get_current_log(jobid)
            error_report = response.json()
            error_report["details"] = log[0]
            raise SolverError(error_report)
        if response.status_code < 499:
            error_report = response.json()
            raise RuntimeError(error_report)
        return ""

    def get_current_status(self, jobid: str) -> str:
        params = {"jobid": str(jobid)}
        response = self.session.get(
            self.server + Endpoints.getcurstatus,
            params=params,
            headers={"X-api-key": self.api_key, **self.custom_headers},
        )

        return json.loads(response.text)

    def get_current_solution(self, jobid: str) -> str:
        params = {"jobid": str(jobid)}
        response = self.session.get(
            self.server + Endpoints.getcursolution,
            params=params,
            headers={"X-api-key": self.api_key, **self.custom_headers},
        )

        array = json.loads(response.text)
        for e in array:
            self._replace_file_content_from_url(e, "solution")
        return array

    def get_current_log(self, jobid: str) -> str:
        params = {"jobid": str(jobid)}
        response = self.session.get(
            self.server + Endpoints.getcurlog, params=params, headers={"X-api-key": self.api_key, **self.custom_headers}
        )
        if not response.ok:
            error_report = response.json()
            raise RuntimeError(error_report)
        return [self._get_file_content_from_url(e) for e in json.loads(response.text)]

    def get_results(self, jobid: str) -> Tuple[dict, int]:
        params = {"jobid": str(jobid)}
        response = self.session.get(
            self.server + Endpoints.getresults,
            params=params,
            headers={"X-api-key": self.api_key, **self.custom_headers},
        )

        if not response.ok:
            error_report = response.json()
            raise RuntimeError(error_report)

        result = json.loads(response.text)
        array = copy.deepcopy(result["result"])

        for e in array:
            self._replace_file_content_from_url(e, "solution_file")
            self._replace_file_content_from_url(e, "solver_log")

        return array, int(result["time_billed"])

    def get_jobs(self, n: int) -> dict:
        params = {"n": str(n)}
        response = self.session.get(
            self.server + Endpoints.getjobs, params=params, headers={"X-api-key": self.api_key, **self.custom_headers}
        )

        if not response.ok:
            error_report = response.json()
            raise RuntimeError(error_report)

        return json.loads(response.text)

    def interrupt_job(self, jobid: str) -> bool:
        response = self.session.delete(
            self.server + Endpoints.interruptjob + "/" + str(jobid),
            headers={"X-api-key": self.api_key, **self.custom_headers},
        )

        if not response.ok:
            error_report = response.json()
            raise RuntimeError(error_report)

==> quantagonia-0.2.0-py3-none-any.whl: quantagonia/cloud/https_client.py <==
from __future__ import annotations

import copy
import json
import os
import uuid

import requests
from requests.adapters import HTTPAdapter
from urllib3 import Retry

from quantagonia.cloud.dto.encoder import DtoJSONEncoder
from quantagonia.cloud.dto.job import JobDto
from quantagonia.cloud.dto.presigned_s3 import PresignedS3
from quantagonia.cloud.enums import Endpoints
from quantagonia.enums import HybridSolverServers
from quantagonia.errors import SolverError


class HTTPSClient:
    """HTTPS Client for the HybridSolver with low-level functions to manage optimization tasks.

    Args:
        api_key (str): The API key to be used to authenticate with the HybridSolver.
        server (HybridSolverServers): (optional) Allows to specify a server different to the default production server.
                                      Defaults to :code:`None`.
        custom_headers (dict): (optional) Custom headers for the https requests.

    """

    def __init__(
        self, api_key: str, server: HybridSolverServers | None = None, custom_headers: dict | None = None
    ) -> None:
        self.api_key = api_key
        if server is None:
            # Read the default server from an env variable. If none is set, the server defaults to PROD
            server = HybridSolverServers[os.environ.get("QUANTAGONIA_SERVER", "PROD").upper()]
        self.server: str = server.value
        self.custom_headers = custom_headers if custom_headers is not None else {}

        self.retry = Retry(total=5, backoff_factor=1, allowed_methods=frozenset(["GET", "POST", "PUT", "DELETE"]))
        self.session = requests.Session()
        self.session.mount("http://", HTTPAdapter(max_retries=self.retry))
        self.session.mount("https://", HTTPAdapter(max_retries=self.retry))

    def submit_job(
        self, problem_files: list, specs: list, tag: str = "", context: str = "", origin: str = "python-api"
    ) -> str:
        # build a single JSON array with specs
        spec_arr = {}
        for ix in range(len(specs)):
            spec_arr[str(ix)] = specs[ix]

        job_id = str(uuid.uuid4())
        problem_filenames = []
        for ind, prob_file in enumerate(problem_files):
            try:
                filename = prob_file.parts[len(prob_file.parts) - 1]
            except AttributeError:
                filename = prob_file.split("/")[len(prob_file.split("/")) - 1]

            problem_filenames.append(filename)
            ps3_problem_file = PresignedS3(
                jobId=job_id, contentType="application/octet-stream", batchNumber=str(ind), fileName=filename
            )
            data_for_s3 = json.dumps(ps3_problem_file, cls=DtoJSONEncoder)
            response = self.session.post(
                self.server + Endpoints.s3,
                data=data_for_s3,
                headers={"X-api-key": self.api_key, "Content-type": "application/json", **self.custom_headers},
            )
            if not response.ok:
                msg = "Unable to get an S3 presigned URL"
                raise RuntimeError(msg)

            file_uploader = response.json()
            headers = {
                "X-amz-meta-author": file_uploader["metaAuthor"],
                "X-amz-meta-version": file_uploader["metaVersion"],
                "Content-type": "application/octet-stream",
            }
            with open(prob_file, "rb") as f:
                upload_response = self.session.put(file_uploader["url"], data=f, headers=headers)

            if not upload_response.ok:
                raise RuntimeError(upload_response.text)

        start_job = JobDto(
            jobId=job_id, problemFiles=problem_filenames, specs=specs, tag=tag, context=context, origin=origin
        )
        start_job_data = json.dumps(start_job, cls=DtoJSONEncoder)
        started_job = self.session.post(
            self.server + Endpoints.job,
            data=start_job_data,
            headers={"X-api-key": self.api_key, "Content-type": "application/json", **self.custom_headers},
        )
        if not started_job.ok:
            error_report = started_job.json()
            raise RuntimeError(error_report)

        return started_job.json()["jobId"]

    def _replace_file_content_from_url(self, e: dict, key: str) -> None:
        e[key] = self._get_file_content_from_url(e[key])

    def _get_file_content_from_url(self, e: dict | str) -> str:
        if isinstance(e, dict) and "error" in e:
            return "Error: " + e["error"]
        if isinstance(e, dict) and "url" in e:
            if e["url"] == "":
                return ""
            response = self.session.get(e["url"])
            if response.status_code == 200:
                return response.text
            return ""
        return e

    def check_job(self, jobid: str) -> str:
        params = {"jobid": str(jobid)}
        response = self.session.get(
            self.server + Endpoints.checkjob, params=params, headers={"X-api-key": self.api_key, **self.custom_headers}
        )
        if response.ok:
            return response.json()["status"]
        if response.status_code > 499:
            log = self.get_current_log(jobid)
            error_report = response.json()
            error_report["details"] = log[0]
            raise SolverError(error_report)
        if response.status_code < 499:
            error_report = response.json()
            raise RuntimeError(error_report)
        error_message = "Unknown error"
        raise RuntimeError(error_message)

    def get_current_status(self, jobid: str) -> str:
        params = {"jobid": str(jobid)}
        response = self.session.get(
            self.server + Endpoints.getcurstatus,
            params=params,
            headers={"X-api-key": self.api_key, **self.custom_headers},
        )

        return json.loads(response.text)

    def get_current_solution(self, jobid: str) -> list[str]:
        params = {"jobid": str(jobid)}
        response = self.session.get(
            self.server + Endpoints.getcursolution,
            params=params,
            headers={"X-api-key": self.api_key, **self.custom_headers},
        )

        if response.ok:
            array = json.loads(response.text)
            if not isinstance(array, list):
                error_message = f"Expected a list from the /getcurrentsolution endpoint, got {type(array)}"
                raise RuntimeError(error_message)
            for batch_item in array:
                self._replace_file_content_from_url(batch_item, "solution")
            return array

        error_report = response.json()
        raise RuntimeError(error_report)

    def get_current_log(self, jobid: str) -> list[str]:
        params = {"jobid": str(jobid)}
        response = self.session.get(
            self.server + Endpoints.getcurlog, params=params, headers={"X-api-key": self.api_key, **self.custom_headers}
        )
        if not response.ok:
            error_report = response.json()
            raise RuntimeError(error_report)
        return [self._get_file_content_from_url(e) for e in json.loads(response.text)]

    def get_results(self, jobid: str) -> tuple[dict, int]:
        params = {"jobid": str(jobid)}
        response = self.session.get(
            self.server + Endpoints.getresults,
            params=params,
            headers={"X-api-key": self.api_key, **self.custom_headers},
        )

        if not response.ok:
            error_report = response.json()
            raise RuntimeError(error_report)

        result = json.loads(response.text)
        array = copy.deepcopy(result["result"])

        for e in array:
            self._replace_file_content_from_url(e, "solution_file")
            self._replace_file_content_from_url(e, "solver_log")

        return array, int(result["time_billed"])

    def get_jobs(self, n: int) -> dict:
        params = {"n": str(n)}
        response = self.session.get(
            self.server + Endpoints.getjobs, params=params, headers={"X-api-key": self.api_key, **self.custom_headers}
        )

        if not response.ok:
            error_report = response.json()
            raise RuntimeError(error_report)

        return json.loads(response.text)

    def interrupt_job(self, jobid: str) -> None:
        response = self.session.delete(
            self.server + Endpoints.interruptjob + "/" + str(jobid),
            headers={"X-api-key": self.api_key, **self.custom_headers},
        )

        if not response.ok:
            error_report = response.json()
            raise RuntimeError(error_report)
